@@ -26,6 +26,8 @@ class Exploration:
         default_params=None,
         exploration_name=None,
         hdf_filename=None,
+        num_cpus: int = None,
+        num_gpus: int = None,
     ):
         """Defines a parameter exploration of a given `function`.
         
@@ -39,7 +41,10 @@ class Exploration:
         :type exploration_name: str, optional
         :param hdf_filename: Filename of the hdf storage file, defaults to None
         :type hdf_filename: str, optional
-        :param full_params: Pass full parameter dict to evaluation function if `True`, or else pass only the explored parameters, defaults to True 
+        :param num_cpus: Number of desired CPU cores passed to ray, defaults to None
+        :type num_cpus: int, optional
+        :param num_gpus: Number of desired GPUs passed to ray, defaults to None
+        :type num_gpus: int, optional
         :return: Exploration instance
         """
 
@@ -66,19 +71,33 @@ class Exploration:
             hdf_filename = "exploration.h5"
         self.hdf_filename = hdf_filename
 
-        self.run_id = None
         self.dfResults = None
 
         # status
         self._hdf_open_for_reading = False
 
+        # List of all parameter combinations generated when exploration starts
+        self.explore_params_list = None
+
+        # Dict with runId as keys and explored parameter dict as value.
+        # Will be filled when exploration starts.
+        self.run_params_dict = {}
+
+        # Dict with runId as keys and explored parameter dict as value.
+        # Will be filled when calling `load_results`.
+        self.params = {}
+
+        # Ray configuration
+        self.num_gpus = num_gpus
+        self.num_cpus = num_cpus
+
     def run(self):
         """Start parameter exploration.
 
         TODO: Pass kwargs in run() to the exploration function
         """
         # Initialize ray
-        self._init_ray()
+        self._init_ray(num_cpus=self.num_cpus, num_gpus=self.num_gpus)
 
         # Create a list of all combinations of parameters from explore_params
         self.explore_params_list = self._cartesian_product_dict(self.explore_params)
@@ -90,7 +109,7 @@ class Exploration:
         # remember the time
         start_time = time.time()
         # a unique id for each run
-        self.run_id = 0
+        run_id = 0
 
         # contains ray objects of each run
         ray_returns = {}
@@ -110,13 +129,13 @@ class Exploration:
 
             # start all ray jobs and remember the ray object
             # pylint: disable=no-member
-            ray_returns[self.run_id] = self._ray_remote.remote(self, run_params)
+            ray_returns[run_id] = self._ray_remote.remote(self, run_params)
 
             # store this runs explore parameters
-            self.run_params_dict[self.run_id] = copy.deepcopy(update_params)
+            self.run_params_dict[run_id] = copy.deepcopy(update_params)
 
             # increment the run id
-            self.run_id += 1
+            run_id += 1
 
         # stop measuring time
         end_time = time.time() - start_time
@@ -138,6 +157,8 @@ class Exploration:
             run_param = self.run_params_dict[run_id]
             # queue object for storage
             self._store_result(run_id, ray_return, run_param)
+            # remove LOCAL_REFERENCE in form of ObjectId from ray's object store.
+            ray_returns[run_id] = None
 
         # stop measuring time
         end_time = time.time() - start_time
@@ -151,12 +172,12 @@ class Exploration:
     def load_results(
         self, filename=None, exploration_name=None, aggregate=True, all=False
     ):
-        """Load results from previouis explorations. This function 
+        """Load results from previous explorations. This function 
         will open an HDF file and look for an exploration. It will 
-        create a Pandas `Dataframe` object (accessible thorugh the 
-        attribute `.df`) with a list of all runs and their paramters. 
+        create a Pandas `Dataframe` object (accessible through the
+        attribute `.df`) with a list of all runs and their parameters.
         
-        You can load the exploration results using following paramters:
+        You can load the exploration results using following parameters:
 
         - If `aggregate==True`, all scalar results (such as `float` 
         or `int`) from the exploration will be added to the Dataframe.
@@ -167,11 +188,11 @@ class Exploration:
         you might want to skip this and load results separately using the
         method `.get_run()`.
 
-        :param filename: Filename of HDF file, usese default filename or previously used filename if not given, defaults to None
+        :param filename: Filename of HDF file, uses default filename or previously used filename if not given, defaults to None
         :type filename: str, optional
         :param exploration_name: Name of the exploration, same as the group names of the explorations in the HDF file, defaults to None
         :type exploration_name: str, optional
-        :param aggregate: Aggregate scalar results into the results Dataframe. If this option is enabled, , defaults to True
+        :param aggregate: Aggregate scalar results into the results Dataframe. If this option is enabled, defaults to True
         :type aggregate: bool, optional
         :param all: Load all results into a dictionary available as the attribute `.results`. Can use a lot of RAM, defaults to False
         :type all: bool, optional
@@ -189,7 +210,7 @@ class Exploration:
         self.close_hdf()
 
     def get_run(self, run_id=None, run_name=None, filename=None, exploration_name=None):
-        """Get a sigle result from a previous exploration. This function
+        """Get a single result from a previous exploration. This function
         will load a single result from the HDF file. Use this function
         if you want to avoid loading all results to memory, which you can 
         do using `.load_results(all=True)`.
@@ -197,14 +218,14 @@ class Exploration:
         Note: This function will open the HDF for reading but will not close
         it afterwards! This is to speed up many sequential loads but it also
         means that you have to close the HDF file yourself. You can do this
-        by uysing `.close_hdf()`.
+        by using `.close_hdf()`.
 
         :param run_id: Unique id of the run. Has to be given if run_name is not given, defaults to None
         :type run_id: int, optional
         :param run_name: The name of the run. Has to be given if run_id is not given, defaults to None
         :type run_name: str, optional
-        :param exploration_name: Filename of the HDF with previous exploration results. Previously used filename will be used if not given, defaults to None
-        :type exploration_name: str, optional
+        :param filename: Filename of the HDF with previous exploration results. Previously used filename will be used if not given, defaults to None
+        :type filename: str, optional
         :param exploration_name: Name of the exploration to load data from. Previously used exploration_name will be used if not given, defaults to None
         :type exploration_name: str, optional
         
@@ -216,9 +237,7 @@ class Exploration:
             run_id is not None or run_name is not None
         ), "Either use `run_id` or `run_name`."
 
-        if exploration_name is None:
-            exploration_name = self.exploration_name
-        else:
+        if exploration_name:
             self.exploration_name = exploration_name
 
         if run_id is not None:
@@ -261,20 +280,25 @@ class Exploration:
         r = self.function(params)
         return r
 
-    def _init_ray(self):
+    def _init_ray(self, num_cpus: int = None, num_gpus: int = None):
         """Initialize ray.
+
+        :param num_cpus: Number of desired CPU cores used for this run, defaults to None
+        :type num_cpus: int, optional
+        :param num_gpus: Number of desired GPUs used for this run, defaults to None
+        :type num_gpus: int, optional
         """
         if ray.is_initialized():
             self._shutdown_ray()
 
-        ray.init()
-        assert ray.is_initialized() == True, "Could not initialize ray."
+        ray.init(num_cpus=num_cpus, num_gpus=num_gpus)
+        assert ray.is_initialized() is True, "Could not initialize ray."
 
     def _shutdown_ray(self):
         """Shutdown ray.
         """
         ray.shutdown()
-        assert ray.is_initialized() == False, "Could not shutdown ray."
+        assert ray.is_initialized() is False, "Could not shutdown ray."
 
     ##############################################
     ## DATA STORAGE
@@ -285,14 +309,16 @@ class Exploration:
         
         :param group: group in hdf file to store data in
         :type group: [type]
-        :param dict: dictionary with data to store
-        :type dict: dict
+        :param dict_data: dictionary with data to store
+        :type dict_data: dict
         """
-        for rkey, rval in dict_data.items():
+        for r_key, r_val in dict_data.items():
             try:
-                self.h5file.create_array(group, rkey, obj=rval)
+                self.h5file.create_array(group, r_key, obj=r_val)
             except:
-                logging.warn(f"Could not store dict entry {rkey} (type: {type(rval)})")
+                logging.warning(
+                    f"Could not store dict entry {r_key} (type: {type(r_val)})"
+                )
 
     def _init_hdf(self):
         """Create hdf storage file and all necessary groups.
@@ -340,8 +366,8 @@ class Exploration:
         :type result_id: int
         :param ray_object: ray object
         :type ray_object: ray object
-        :param run_param: explored parameters of the run
-        :type run_param: dict
+        :param run_params: explored parameters of the run
+        :type run_params: dict
         """
         # set the name of this run for naming the hdf group
         run_result_name = self.RUN_PREFIX + str(result_id)
@@ -354,16 +380,12 @@ class Exploration:
         ), f"Returned result must be a dictionary, is `{type(result_dict)}`."
 
         self._store_result_in_hdf(run_result_name, result_dict, run_params)
-        # store all results in a dictionary
-        # self._store_result_in_dictionary(result_id, result_dict)
 
     def _store_result_in_hdf(self, run_result_name, result_dict, run_params):
         """Stores the results of a ray object of a single run and the parameters of the run.
         
         :param run_result_name: Name of the result
         :type run_result_name: str
-        :param run_return: ray object of the run
-        :type run_return: ray object
         :param run_params: Explored parameters of the run
         :type run_params: dict
         """
@@ -371,8 +393,8 @@ class Exploration:
         # create the results group
         run_results_group = self.h5file.create_group(self.runs_group, run_result_name)
         # store each item in the dictionary
-        for rkey, rval in result_dict.items():
-            self.h5file.create_array(run_results_group, rkey, obj=rval)
+        for r_key, r_val in result_dict.items():
+            self.h5file.create_array(run_results_group, r_key, obj=r_val)
 
         # store parameters
         # create parameters group
@@ -380,14 +402,16 @@ class Exploration:
         # store the parameter dictionary
         self._store_dict_to_hdf(run_params_group, run_params)
 
-    def _store_result_in_dictionary(self, result_id, result_dict):
-        self.results[self.run_id] = copy.deepcopy(result_dict)
-
     ##############################################
     ## READ DATA
     ##############################################
 
     def _create_df(self):
+        """Create results dataframe and will it with columns according to the explored parameters.
+
+        :return: dfResults
+        :rtype: pd.DataFrame
+        """
         logging.info("Creating new results DataFrame")
         self.explore_params = self._read_explore_params()
         self.dfResults = pd.DataFrame(
@@ -398,6 +422,11 @@ class Exploration:
         return self.dfResults
 
     def _open_hdf(self, filename=None):
+        """Open a hdf file as `.h5file` for reading later.
+
+        :param filename: Filename of HDF file, defaults to None
+        :type filename: str, optional
+        """
         if filename is not None:
             self.hdf_filename = filename
         assert (
@@ -416,6 +445,14 @@ class Exploration:
         logging.info(f"{self.hdf_filename} closed.")
 
     def _aggregate_results(self, exploration_name=None):
+        """Go through all results saved in `.results` and store all floats in the results table.
+
+        TODO: Direct reading from hdf without having to load it to memory, like in neurolib
+        TODO: Storage of non-scalar values, like in neurolib
+
+        :param exploration_name: [description], defaults to None
+        :type exploration_name: [type], optional
+        """
         nan_value = np.nan
         logging.info("Aggregating scalar results ...")
         for runId, parameters in tqdm.tqdm(
@@ -431,6 +468,13 @@ class Exploration:
         self.dfResults = self.dfResults.dropna(axis="columns", how="all")
 
     def _load_all_results(self, exploration_name=None, all=True):
+        """Load all results in hdf file into `.results` (if all=True). Load all explored parameters into `.params`.
+
+        :param exploration_name: Name of the run, defaults to None
+        :type exploration_name: str, optional
+        :param all: Whether to load everything into ram or not, defaults to True
+        :type all: bool, optional
+        """
         if exploration_name is None:
             exploration_name = self.exploration_name
         else:
@@ -461,6 +505,13 @@ class Exploration:
         logging.debug(f"{len(self.results)} results loaded to memory.")
 
     def _read_group_as_dict(self, group):
+        """Read an HDF group and return a dictionary.
+
+        :param group: hdf group
+        :type group: tables.group.Group
+        :return: Dictionary with data of the group.
+        :rtype: dict
+        """
         return_dict = {}
         # iterate through all arrays in that group
         for array in group:
@@ -476,6 +527,13 @@ class Exploration:
         return return_dict
 
     def _get_run_parameters(self, hdf_run):
+        """Get the parameters of a run
+
+        :param hdf_run: Tables group
+        :type hdf_run: tables.group.Group
+        :return: Parameters as a dictionary
+        :rtype: dict
+        """
         run_params = {}
         if isinstance(hdf_run, tables.group.Group):
             run_params_group = self.h5file.get_node(hdf_run, "params")
@@ -483,6 +541,11 @@ class Exploration:
         return run_params
 
     def _read_explore_params(self):
+        """Get explored parameters of from opened hdf file.
+
+        :return: Dictionary with explored parameters
+        :rtype: dict
+        """
         explore_params_group = self.h5file.get_node(
             "/" + self.exploration_name, "explore_params"
         )
@@ -491,6 +554,11 @@ class Exploration:
 
     @property
     def df(self):
+        """Returns a dataframe with exploration results. Creates it from new if it doesn't exist yet.
+
+        :return: Dataframe with exploration results
+        :rtype: pandas.DataFrame
+        """
         if hasattr(self, "dfResults"):
             if self.dfResults is None:
                 return self._create_df()
